#include "wtile.hh"
#include "matlab_fcs.hh"
#include <constant.hh>
#include "DVecType.hh"
#include "Bits.hh"
#include "fSeries/DFT.hh"
#include <cmath>
#include <iostream>

using namespace std;
using namespace wpipe;
using namespace containers;

//
//    Notes on variables used here:
//    timeRange: Length of time series that will be ffted
//    minimumFrequencyStep: frequency step in dft of time series
//

//======================================  wtile constructor.
wtile::wtile(double timeRange, const dble_vect& qRange, 
	     const dble_vect& frequencyRange, double sampleFrequency, 
	     double maximumMismatch, double highPassCutoff, 
	     double lowPassCutoff, double whiteningDuration, 
	     double transientFactor) {

  const double dInf = 1.0 / 0.0;
  // apply default arguments

  // extract minimum and maximum Q from Q range
  double minimumQ = qRange[0];
  double maximumQ = qRange[1];

  // extract minimum and maximum frequency from frequency range
  double minimumFrequency = 0;
  double maximumFrequency = dInf;
  if (!frequencyRange.empty()) {
    minimumFrequency = frequencyRange[0];
    if (frequencyRange.size() == 1) maximumFrequency = frequencyRange[0];
    else                            maximumFrequency = frequencyRange[1];
  }

  //////////////////////////////////////////////////////////////////////////////
  //                          compute derived parameters                      //
  //////////////////////////////////////////////////////////////////////////////

  // nyquist frequency
  double nyquistFrequency = sampleFrequency / 2;

  // maximum mismatch between neighboring tiles
  double mismatchStep = 2 * sqrt(maximumMismatch / 3);

  // maximum possible time resolution
  //double minimumTimeStep = 1 / sampleFrequency;

  // maximum possible frequency resolution
  double minimumFrequencyStep = 1 / timeRange;

  // conversion factor from Q prime to true Q
  double qPrimeToQ = sqrt(11);

  // total number of samples in input data
  double numberOfSamples = timeRange * sampleFrequency;

  /////////////////////////////////////////////////////////////////////////////
  //                       determine parameter constraints                   //
  /////////////////////////////////////////////////////////////////////////////

  // minimum allowable Q prime to prevent window aliasing at zero frequency
  double minimumAllowableQPrime = 1.0;

  // minimum allowable Q to avoid window aliasing at zero frequency
  double minimumAllowableQ = minimumAllowableQPrime * qPrimeToQ;

  // reasonable number of statistically independent tiles in a frequency row
  double minimumAllowableIndependents = 50;

  // maximum allowable mismatch parameter for reasonable performance
  double maximumAllowableMismatch = 0.5;

  /////////////////////////////////////////////////////////////////////////////
  //                             validate parameters                         //
  /////////////////////////////////////////////////////////////////////////////

  // check for valid time range
  if (timeRange < 0) {
    error("negative time range");
  }

  // check for valid Q range
  if (minimumQ > maximumQ) {
    error("minimum Q exceeds maximum Q");
  }

  // check for valid frequency range
  if (minimumFrequency > maximumFrequency) {
    error("minimum frequency exceeds maximum frequency");
  }

  // check for valid minimum Q
  if (minimumQ < minimumAllowableQ) {
    error("minimum Q less than minimum allowable");
  }

  // check for reasonable maximum mismatch parameter
  if ( maximumMismatch > maximumAllowableMismatch) {
    error("maximum mismatch exceeds maximum allowable");
  }

  // check for integer power of two data length
  //if (fmod(log(timeRange * sampleFrequency) / log(2), 1) != 0) {
  if (!is_power_of_2(timeRange * sampleFrequency)) {
    error("data length is not an integer power of two");
  }

  //////////////////////////////////////////////////////////////////////////////
  //                              determine Q planes                          //
  //////////////////////////////////////////////////////////////////////////////

  // cumulative mismatch across Q range
  double qCumulativeMismatch = log(maximumQ / minimumQ) / sqrt(2);

  // number of Q planes
  int numberOfPlanes = int(ceil(qCumulativeMismatch / mismatchStep));

  // insure at least one plane
  if (numberOfPlanes == 0) {
    numberOfPlanes = 1;
  }

  // mismatch between neighboring planes
  double qMismatchStep = qCumulativeMismatch / numberOfPlanes;

  // index of Q planes
  // double qIndices = 0.5 : numberOfPlanes - 0.5;

  // vector of Qs
  //double qs = minimumQ * exp(sqrt(2) * qIndices * qMismatchStep);
  double minQ = minimumQ * exp(sqrt(2) * 0.5 * qMismatchStep);
  double maxQ = minimumQ * exp(sqrt(2) * (numberOfPlanes-0.5) * qMismatchStep);

  //////////////////////////////////////////////////////////////////////////////
  //                             validate frequencies                         //
  //////////////////////////////////////////////////////////////////////////////

  // minimum allowable frequency to provide sufficient statistics
  double minimumAllowableFrequency = minimumAllowableIndependents * maxQ / 
    (2 * pi * timeRange);

  // maximum allowable frequency to avoid window aliasing
  double maximumAllowableFrequency = nyquistFrequency / (1 + qPrimeToQ / minQ);

  // check for valid minimum frequency
  if (maximumFrequency < 0.0) maximumFrequency = dInf;
  if ((minimumFrequency != 0 && minimumFrequency < minimumAllowableFrequency) ||
      (maximumFrequency != dInf && maximumFrequency > maximumAllowableFrequency)) { 
    cerr << "Requested frequency range (" << minimumFrequency << "-" 
	 << maximumFrequency <<") is not within allowable range (" 
	 << minimumAllowableFrequency << "-" << maximumAllowableFrequency
	 << ")" << endl;
     error("Requested frequency range is not within allowable range");
  }

  //////////////////////////////////////////////////////////////////////////////
  //                     create Q transform tiling structure                  //
  //////////////////////////////////////////////////////////////////////////////

  // structure type identifier
  _id = "Discrete Q-transform tile structure";

  // insert duration into tiling structure
  _duration = timeRange;

  // insert minimum Q into tiling structure
  _minimumQ = minimumQ;

  // insert maximum Q into tiling structure
  _maximumQ = maximumQ;

  // insert minimum frequency into tiling structure
  _minimumFrequency = minimumFrequency;

  // insert maximum frequency into tiling structure
  _maximumFrequency = maximumFrequency;

  // insert sample frequency into tiling structure
  _sampleFrequency = sampleFrequency;

  // insert maximum loss due to mismatch into tiling structure
  _maximumMismatch = maximumMismatch;

  // insert Q vector into tiling structure
  // _qs = qs;

  // insert number of Q planes into tiling structure
  _numberOfPlanes = numberOfPlanes;

  // initialize cell array of Q plans in tiling structure
  //_planes = cell(1, numberOfPlanes);
  _planes.resize(numberOfPlanes);

  // initialize total number of tiles counter
  _numberOfTiles = 0;

  // initialize total number of independent tiles counter
  _numberOfIndependents = 0;

  // initialize total number of flops counter
  _numberOfFlops = double(numberOfSamples) * log(double(numberOfSamples));

  //////////////////////////////////////////////////////////////////////////////
  //                           begin loop over Q planes                       //
  //////////////////////////////////////////////////////////////////////////////

  // begin loop over Q planes
  for (int plane=0; plane < numberOfPlanes; plane++) {

    // calculate Q of plane
    double q = minimumQ * exp(sqrt(2) * (plane + 0.5) * qMismatchStep);

    ////////////////////////////////////////////////////////////////////////////
    //                        determine plane properties                      //
    ////////////////////////////////////////////////////////////////////////////

    // find Q prime for the plane
    double qPrime = q / qPrimeToQ;

    // for large qPrime use asymptotic value of planeNormalization
    double planeNormalization = 1;
    
    // otherwise
    if (qPrime <= 10) {
      // polynomial coefficients for plane normalization factor
      double coefficients[9];
      coefficients[0] = + 1. * log((qPrime + 1) / (qPrime - 1));
      coefficients[1] = - 2.; 
      coefficients[2] = - 4. * log((qPrime + 1) / (qPrime - 1));
      coefficients[3] =  22.0 / 3.0; 
      coefficients[4] = + 6. * log((qPrime + 1) / (qPrime - 1));
      coefficients[5] = - 146.0 / 15.0; 
      coefficients[6] = - 4. * log((qPrime + 1) / (qPrime - 1));
      coefficients[7] = + 186.0 / 35.0;
      coefficients[8] = + 1. * log((qPrime + 1) / (qPrime - 1));

      // plane normalization factor
      double pval = polyval(coefficients, 9, qPrime);
      planeNormalization = sqrt(256.0 / (315.0 * qPrime * pval));


      // continue
    }

    ///////////////////////////////////////////////////////////////////////////
    //                        determine frequency rows                       //
    ///////////////////////////////////////////////////////////////////////////

    // minimum allowable frequency to provide sufficient statistics
    double minimumAllowableFrequency = minimumAllowableIndependents*q / 
      (2*pi * _duration);

    // plane specific maximum allowable frequency to avoid window aliasing
    double maximumAllowableFrequency = nyquistFrequency / (1 + qPrimeToQ / q);

    // use plane specific minimum allowable frequency if (requested
    if (_minimumFrequency == 0) {
      minimumFrequency = minimumAllowableFrequency;
    }

    // use plane specific maximum allowable frequency if (requested
    if (_maximumFrequency < 0 || _maximumFrequency == dInf) {
      maximumFrequency = maximumAllowableFrequency;
    }

    // cumulative mismatch across frequency range
    double frequencyCumulativeMismatch = 
           log(maximumFrequency / minimumFrequency) * sqrt(2 + q*q) / 2;

    // number of frequency rows
    int numberOfRows = int(ceil(frequencyCumulativeMismatch / mismatchStep));

    // insure at least one row
    if (numberOfRows == 0) {
      numberOfRows = 1;
    }

    // mismatch between neighboring frequency rows
    double frequencyMismatchStep = frequencyCumulativeMismatch / numberOfRows;

    // index of frequency rows
    //frequencyIndices = 0.5 : numberOfRows - 0.5;
    //
    // vector of frequencies
    //frequencies = minimumFrequency * exp((2 / sqrt(2 + q^2)) * 
    //					 frequencyIndices * 
    //					 frequencyMismatchStep);
    //
    // project frequency vector onto realizable frequencies
    //   frequencies = round(frequencies / minimumFrequencyStep) .* 
    //  minimumFrequencyStep;

    // ratio between successive frequencies
    double frequencyRatio = exp((2 / sqrt(2 + q*q)) * frequencyMismatchStep);

    ///////////////////////////////////////////////////////////////////////////
    //                   create Q transform plane structure                  //
    ///////////////////////////////////////////////////////////////////////////

    // insert Q of plane into Q plane structure
    _planes[plane].q = q;

    // insert minimum search frequency of plane into Q plane structure
    _planes[plane].minimumFrequency = minimumFrequency;

    // insert maximum search frequency of plane into Q plane structure
    _planes[plane].maximumFrequency = maximumFrequency;

    // insert plane normalization factor into Q plane structure
    _planes[plane].normalization = planeNormalization;

    // insert frequency vector into Q plane structure
    //_planes[plane].frequencies = frequencies;

    // insert number of frequency rows into Q plane structure
    _planes[plane].numberOfRows = numberOfRows;

    // initialize cell array of frequency rows into Q plane structure
    _planes[plane].rows.resize(numberOfRows);

    // initialize number of tiles in plane counter
    _planes[plane].numberOfTiles = 0;

    // initialize number of independent tiles in plane counter
    _planes[plane].numberOfIndependents = 0;

    // initialize number of flops in plane counter
    _planes[plane].numberOfFlops = 0;

    ///////////////////////////////////////////////////////////////////////////
    //                      begin loop over frequency rows                   //
    ///////////////////////////////////////////////////////////////////////////

    // begin loop over frequency rows
    for (int row=0; row < numberOfRows; row++) {

      // extract frequency of row from frequency vector
      //frequency = frequencies(row);
      double frequency = minimumFrequency * exp((2 / sqrt(2 + q*q)) * 
						double(row + 0.5) * 
						frequencyMismatchStep);
      frequency=long(frequency/minimumFrequencyStep+0.5)*minimumFrequencyStep;

      /////////////////////////////////////////////////////////////////////////
      //                      determine tile properties                      //
      /////////////////////////////////////////////////////////////////////////

      // bandwidth for coincidence testing
      double bandwidth = 2 * sqrt(pi) * frequency / q;

      // duration for coincidence testing
      double duration = 1 / bandwidth;

      // frequency step for integration
      double frequencyStep = frequency * (frequencyRatio - 1) 
	                   / sqrt(frequencyRatio);

      /////////////////////////////////////////////////////////////////////////
      //                         determine tile times                        //
      /////////////////////////////////////////////////////////////////////////

      // cumulative mismatch across time range
      double timeCumulativeMismatch = timeRange * 2 * pi * frequency / q;

      // number of time tiles
      long logNumberOfTiles = nextpow2(timeCumulativeMismatch/mismatchStep);
      long numberOfTiles = 1 << logNumberOfTiles;

      // mismatch between neighboring time tiles
      double timeMismatchStep = timeCumulativeMismatch / numberOfTiles;

      // index of time tiles
      //timeIndices = 0 : numberOfTiles - 1;

      // time step for integration
      double timeStep = q * timeMismatchStep / (2 * pi * frequency);

      // vector of times
      //times = timeStep * timeIndices;

      // project time vector onto realizable times
      // times = round(times / minimumTimeStep) .* minimumTimeStep;

      // number of flops to compute row
      double numberOfFlops = double(numberOfTiles*logNumberOfTiles) * log(2.0);

      // number of independent tiles in row
      double numberOfIndependents = 1 + timeCumulativeMismatch;

      ////////////////////////////////////////////////////////////////////////
      //                           generate window                            //
      ////////////////////////////////////////////////////////////////////////

      // half length of window in samples
      int halfWindowLength = int((frequency / qPrime) / minimumFrequencyStep);

      // full length of window in samples
      int windowLength = 2 * halfWindowLength + 1;

      // sample index vector for window construction
      // windowIndices = -halfWindowLength : halfWindowLength;

      DVectD dvd(windowLength);
      dvd.replace_with_zeros(0, windowLength, windowLength);

      // frequency vector for window construction
      // windowFrequencies = windowIndices * minimumFrequencyStep;

      // dimensionless frequency vector for window construction
      // double windowArgument = windowFrequencies * qPrime / frequency;

      // bi square window function
      // window = (1 - windowArgument.^2).^2;
      dvd[halfWindowLength] = 1.0;
      for (int i=1; i<=halfWindowLength; ++i) {
	double wFreq = minimumFrequencyStep * double(i);
	double wArg  = wFreq * qPrime / frequency;
	double wVal  = pow(1.0 - wArg * wArg, 2);
	dvd[halfWindowLength+i] = wVal;
	dvd[halfWindowLength-i] = wVal;
      }

      // row normalization factor
      double rowNormalization = sqrt((315 * qPrime) / (128 * frequency));

      // inverse fft normalization factor
      double ifftNormalization = numberOfTiles / numberOfSamples;

      // normalize window
      // window = window * ifftNormalization * rowNormalization * 
      //          planeNormalization;
      dvd *= ifftNormalization * rowNormalization;

      //  Build the fSeries
      DFT window;
      static_cast<fSeries&>(window) =
	fSeries(-halfWindowLength * minimumFrequencyStep, minimumFrequencyStep,
		Time(0), 1./minimumFrequencyStep, dvd);

      // number of zeros to append to windowed data
      int zeroPadLength = numberOfTiles - windowLength;

      // vector of data indices to inverse fourier transform
      // i.e. a list of indices for frequencies from f0 - windowBW/dF to
      // dataIndices = round(1 + frequency/minimumFrequencyStep + windowIndices);

      ////////////////////////////////////////////////////////////////////////
      //                   create Q transform row structure                 //
      ////////////////////////////////////////////////////////////////////////

      // insert frequency of row into frequency row structure
      _planes[plane].rows[row].frequency = frequency;

      // insert duration into frequency row structure
      _planes[plane].rows[row].duration = duration;

      // insert bandwidth into frequency row structure
      _planes[plane].rows[row].bandwidth = bandwidth;

      // insert time step into frequency row structure
      _planes[plane].rows[row].timeStep = timeStep;

      // insert frequency step into frequency row structure
      _planes[plane].rows[row].frequencyStep = frequencyStep;

      // insert time vector into frequency row structure
      // _planes[plane].rows[row].times = times;
      // THIS FIELD HAS BEEN REMOVED DUE TO EXCESSVE MEMORY USE
      // WHERE REQUIRED, COMPUTE IT BY:
      //     times = (0 :  _planes[plane].rows[row].numberOfTiles - 1) 
      //       * _planes[plane].rows[row].timeStep;

      // insert window vector into frequency row structure
      _planes[plane].rows[row].window = new containers::DFT(window);

      // insert window vector into frequency row structure
      _planes[plane].rows[row].zeroPadLength = zeroPadLength;

      // insert data index vector into frequency row structure
      //_planes[plane].rows[row].dataIndices = dataIndices;

      // insert number of time tiles into frequency row structure
      _planes[plane].rows[row].numberOfTiles = numberOfTiles;

      // insert number of independent tiles in row into frequency row structure
      _planes[plane].rows[row].numberOfIndependents = numberOfIndependents;

      // insert number of flops to compute row into frequency row structure
      _planes[plane].rows[row].numberOfFlops = numberOfFlops;

      // increment number of tiles in plane counter
      _planes[plane].numberOfTiles += numberOfTiles;

      // increment number of indepedent tiles in plane counter
      _planes[plane].numberOfIndependents += numberOfIndependents * 
	(1 + frequencyCumulativeMismatch) / numberOfRows;

      // increment number of flops in plane counter
      _planes[plane].numberOfFlops += numberOfFlops;

      /////////////////////////////////////////////////////////////////////////
      //                       end loop over frequency rows                  //
      /////////////////////////////////////////////////////////////////////////
    }

    // increment total number of tiles counter
    _numberOfTiles += _planes[plane].numberOfTiles;

    // increment total number of independent tiles counter
    _numberOfIndependents += int(_planes[plane].numberOfIndependents * 
				 (1 + qCumulativeMismatch) / numberOfPlanes);

    // increment total number of flops counter
    _numberOfFlops += _planes[plane].numberOfFlops;

    ///////////////////////////////////////////////////////////////////////////
    //                            end loop over Q planes                     //
    ///////////////////////////////////////////////////////////////////////////
    // end loop over Q planes
  }

  /////////////////////////////////////////////////////////////////////////////
  //                         determine filter properties                     //
  /////////////////////////////////////////////////////////////////////////////
  // default high pass filter cutoff frequency
  double defaultHighPassCutoff = _planes[0].minimumFrequency;
  for (int plane=1; plane<_numberOfPlanes; plane++) {
    defaultHighPassCutoff = min(defaultHighPassCutoff, 
				_planes[plane].minimumFrequency);
  }

  // default low pass filter cutoff frequency
  double defaultLowPassCutoff = 0;
  for (int plane=0; plane<_numberOfPlanes; plane++) {
    defaultLowPassCutoff = max(defaultLowPassCutoff, 
			       _planes[plane].maximumFrequency);
  }
 
  // default whitening filter duration
  double defaultWhiteningDuration = 0;
  for (int plane=0; plane<_numberOfPlanes; plane++) {
    defaultWhiteningDuration = max(defaultWhiteningDuration, 
				   _planes[plane].q / 
				   (2 * _planes[plane].minimumFrequency));
  }

  // high pass filter cutoff frequency
  if (highPassCutoff < 0) {
    _highPassCutoff = defaultHighPassCutoff;
  } else {
    _highPassCutoff = highPassCutoff;
  }

  // low pass filter cutoff frequency
  if (lowPassCutoff < 0) {
    _lowPassCutoff = defaultLowPassCutoff;
  } else {
    _lowPassCutoff = lowPassCutoff;
  }
 
  // whitening filter duration
  if (whiteningDuration <= 0) {
    _whiteningDuration = defaultWhiteningDuration;
  } else {
    _whiteningDuration = whiteningDuration;
  }

  // estimated duration of filter transients to supress
  _transientDuration = transientFactor * _whiteningDuration;

  // test for insufficient data
  if ((2 * _transientDuration) >= _duration) {
    error("duration of filter transients equals or exceeds data duration");
  }

  /////////////////////////////////////////////////////////////////////////////
  //                          return Q transform tiling                      //
  /////////////////////////////////////////////////////////////////////////////
}

wtile::~wtile(void) {
}

size_t
wtile::nearest_plane(double qTest) const {
  size_t pBest = 0;
  if (_numberOfPlanes != 0 && qTest > 0) {
    double dq = abs(log(planes(0).q/qTest));
    for (int i=1; i < _numberOfPlanes; i++) {
      double dqi = abs(log(_planes[i].q/qTest));
      if (dqi < dq) {
	pBest = i;
	dqi = dq;
      }
    }
  }
  return pBest;
}

//======================================  Construct a row structure.
qrow::qrow(void)
  : window(0)
{}

//======================================  Destroy a row structure.
qrow::~qrow(void) {
  delete window;
}


//======================================  display a row structure.
ostream& 
qrow::display(std::ostream& out) const {
  out << "frequency: "     << frequency << endl;
  out << "duration: "      << duration << endl;
  out << "bandwidth: "     << bandwidth << endl;
  out << "timeStep: "      << timeStep << endl;
  out << "frequencyStep: " << frequencyStep << endl;
  out << "times:         " << times << endl;
  out << "window:        " << window->Dump(out) << endl;
  out << "dataIndices ---- Note defined yet" << endl;
  out << "zeroPadLength: " << zeroPadLength << endl;
  out << "numberOfTiles: " << numberOfTiles << endl;
  out << "numberOfIndependents: " << numberOfIndependents << endl;
  out << "numberOfFlops: " << numberOfFlops << endl;
  return out;
}

qplane::qplane(void) {
}

qplane::~qplane(void) {
}

//======================================  display a row structure.
ostream& 
qplane::display(std::ostream& out) const {
  out << "q:             " << q << endl;
  out << "minimumFrequency:     " << minimumFrequency << endl;
  out << "maximumFrequency:     " << maximumFrequency << endl;
  out << "normalization:        " << normalization << endl;
  out << "numberOfRows:         " << numberOfRows << endl;
  for (int i=0; i<numberOfRows; i++) {
    out << "rows[" << i << "]:" << endl;
    rows[i].display(out);
  }
  out << "numberOfTiles:        " << numberOfTiles << endl;
  out << "numberOfIndependents: " << numberOfIndependents << endl;
  out << "numberOfFlops:        " << numberOfFlops << endl;
  return out;
}


//======================================  display a tiling structure.
ostream& 
wtile::display(std::ostream& out) const {
  out << "id:                   " << _id << endl;
  out << "duration:             " << _duration << endl;
  out << "minimumQ:             " << _minimumQ << endl;
  out << "maximumQ:             " << _maximumQ << endl;
  out << "minimumFrequency:     " << _minimumFrequency << endl;
  out << "maximumFrequency:     " << _maximumFrequency << endl;
  out << "sampleFrequency:      " << _sampleFrequency << endl;
  out << "maximumMismatch:      " << _maximumMismatch << endl;
  out << "numberOfPlanes:       " << _numberOfPlanes << endl;
  for (int i=0; i<_numberOfPlanes; i++) {
    out << "planes[" << i << "]:" << endl;
    _planes[i].display(out);
  }
  out << "numberOfTiles:        " << _numberOfTiles << endl;
  out << "numberOfIndependents: " << _numberOfIndependents << endl;
  out << "numberOfFlops:        " << _numberOfFlops << endl;
  out << "highPassCutoff:       " << _highPassCutoff << endl;
  out << "lowPassCutoff:        " << _lowPassCutoff << endl;
  out << "whiteningDuration:    " << _whiteningDuration << endl;
  out << "transientDuration:    " << _transientDuration << endl;
  return out;
}
